/*

			Millennium Screen Saver for Windows

				Image Generation and Display

*/

#include "milscrsv.h"

int ImageWid = -1;				  			// Width of master image
int ImageHgt = -1;							// Height of master image
static HPALETTE imagePal = NULL;			// Custom colour palette

static struct {
	WORD	palVersion;
	WORD	palNumEntries;
	PALETTEENTRY palPalEntry[256];
} logpal;

static HGLOBAL earthBitmap = NULL;			// Copy of bitmap resource

//	PALMAPSTART  --  Create a replacement palette bitmap header.

static HGLOBAL pmBuf = NULL;

static LPBITMAPINFOHEADER palMapStart(LPBITMAPINFOHEADER bh, int *bmode)
{
	WORD i, ncol, offbits;
	LPBITMAPINFOHEADER bm = NULL;
	unsigned short *palidx;
    
    assert(pmBuf == NULL);
	ncol = (WORD) (bh->biClrUsed == 0 ? (1L << bh->biBitCount) : bh->biClrUsed);
	offbits = (WORD) (bh->biSize + ncol * sizeof(RGBQUAD));
	pmBuf = GlobalAlloc(GMEM_MOVEABLE, (DWORD) offbits);
	if (pmBuf != NULL) {
		bm = GlobalLock(pmBuf);
	}
	if (bm != NULL) {
		*bmode = DIB_PAL_COLORS;
		memcpy(bm, bh, offbits);
		bm->biClrUsed = bm->biClrImportant = 256;
		palidx = (unsigned short *) ((LPBITMAPINFO) bm)->bmiColors; 
		for (i = 0; i < bm->biClrUsed; i++) {
			*palidx++ = (unsigned short) i;
		}
		return bm;
	}
	*bmode = DIB_RGB_COLORS;
	return bh;	
}

//	PALMAPEND  --  Release palette map buffer, if any.

static void palMapEnd(void)
{
	if (pmBuf != NULL) {
		GlobalUnlock(pmBuf);
		GlobalFree(pmBuf);
		pmBuf = NULL;
	}
} 

//  NEEDIMAGEBITMAP  --  Make sure image bitmap and palette are loaded.

void needImageBitmap(void)
{
	int i;
	LPBITMAPINFOHEADER bh;
	DWORD bmpsize;
	HANDLE earthImageDIB;			// Handle to image DIB in resource

	ImageWid = ImageHgt = -1;

	earthImageDIB = LoadResource(hMainInstance,
					FindResource(hMainInstance, MAKEINTRESOURCE(IDB_MILLENNIUM), RT_BITMAP));
	if (earthImageDIB == NULL) {
		return;
	}

	bh = (LPBITMAPINFOHEADER) LockResource(earthImageDIB);
	if (bh == NULL) {
		return;
	}
	/*	Isn't it great what you have to do to figure out how long a bitmap
		is in memory, thanks to Monkeysoft's allowing the bitmap size and
		colours-used fields to be left as zero?  */
	bmpsize = bh->biSize +
		((bh->biClrUsed == 0 ? (1L << bh->biBitCount) : bh->biClrUsed) * sizeof(RGBQUAD)) +
		((((bh->biWidth + (sizeof(LONG) - 1)) / sizeof(LONG)) * sizeof(LONG)) *
		 bh->biHeight);

	earthBitmap = GlobalAlloc(GMEM_FIXED, bmpsize);
	if (earthBitmap == NULL) {
		UnlockResource(earthImageDIB);
		FreeResource(earthImageDIB);
		earthImageDIB = NULL;
		return;
	}
	memcpy((LPBYTE) earthBitmap, bh, bmpsize); 
	ImageWid = (int) bh->biWidth;
	ImageHgt = (int) bh->biHeight;
    
    if (imagePal == NULL) {
	
		/* Extract the day and night colour palettes from the DIB file and
		   construct a memory logical palette embodying those colours.  Then
		   register the palette with the system to provide those colours on
		   the display. */
	
		logpal.palVersion = 0x300;
		logpal.palNumEntries = 256;
		for (i = 0; i < 256; i++) {
			RGBQUAD q = *((LPRGBQUAD) &((LPBITMAPINFO) bh)->bmiColors[i]);

			logpal.palPalEntry[i].peRed = q.rgbRed;
			logpal.palPalEntry[i].peGreen = q.rgbGreen;
			logpal.palPalEntry[i].peBlue = q.rgbBlue;
			logpal.palPalEntry[i].peFlags = PC_NOCOLLAPSE;
		}
	
		imagePal = CreatePalette((LPLOGPALETTE) &logpal);
	}
	UnlockResource(earthImageDIB);
	FreeResource(earthImageDIB);
	earthImageDIB = NULL;
}

/*  PAINTDIB  --  Paint the saved texture mapped DIB into the device surface.  */

static void paintDIB(HDC hDC, int offx, int offy, int isize)
{
	LPBITMAPINFOHEADER obmap, bp;
	WORD ncol, offbits;
	int bmode = DIB_RGB_COLORS;
	unsigned char *opixel;
	HPALETTE opal;

	obmap = (LPBITMAPINFOHEADER) earthBitmap;
	ncol = (WORD) (obmap->biClrUsed == 0 ? (1L << obmap->biBitCount) : obmap->biClrUsed);
	bp = palMapStart(obmap, &bmode);
	offbits = (WORD) (obmap->biSize + ncol * sizeof(RGBQUAD));
	opixel = ((unsigned char *) (obmap)) + offbits;
	opal = SelectPalette(hDC, imagePal, FALSE);
	RealizePalette(hDC);

	StretchDIBits(hDC, offx, offy, isize, isize,
		0, 0, ImageWid, ImageHgt,
		opixel, (LPBITMAPINFO) bp, bmode, SRCCOPY);
    palMapEnd();
	SelectPalette(hDC, opal, FALSE);
}

//  UPDVIEWFROM  --  Update view from.

void updviewfrom(HWND hWnd, int isize, int offX, int offY, BOOL repositioned)
{
	HDC hDC;
	HFONT rfont, ofont;
	SIZE sz;
    struct tm lt;
	char s[132];
	double ttg = (targetDate + 0.5) - faketime;
	char *verb;
	int which, howmany, fontSize;
	static int changeMode = -1;
	static struct tm llt;

	if (earthBitmap == NULL || imagePal == NULL) {
#ifdef TRACY
		Blooie("Wavy!  Could not load bitmap");
#endif
		return;									// Couldn't load bitmap
	}

	jyear(faketime, &lt.tm_year, &lt.tm_mon, &lt.tm_mday);
	lt.tm_year -= 1900;
	lt.tm_mon--;
	jhms(faketime, &lt.tm_hour, &lt.tm_min, &lt.tm_sec);
	lt.tm_wday = jwday(faketime);

	/*	If nothing has changed which will affect the image,
		skip updating it.  */

	if (!repositioned && ((changeMode == 1) ||
		((changeMode == 2) && (lt.tm_mday == llt.tm_mday)) ||
		((changeMode == 3) && (lt.tm_hour == llt.tm_hour)) ||
		((changeMode == 4) && (lt.tm_min == llt.tm_min)) ||
		((changeMode == 5) && (lt.tm_sec == llt.tm_sec)))) {
		return;
	}

	hDC = GetDC(hWnd);

	if (repositioned) {
		RECT rc;

		GetClientRect(hWnd, &rc);
		FillRect(hDC, &rc, (HBRUSH) GetStockObject(BLACK_BRUSH));
	}

	llt = lt;
	paintDIB(hDC, offX, offY, isize);

	rfont = CreateFont(40, 0, 0, 0, FW_BOLD, 0, 0, 0, ANSI_CHARSET,
					   OUT_TT_PRECIS, CLIP_TT_ALWAYS, PROOF_QUALITY,
					   0x04 | FF_ROMAN, NULL);
	ofont = SelectObject(hDC, rfont);

	ttg -= 1.0e-6;
	if (ttg <= 0) {
		lstrcpy(s, Format(7));
		if (changeMode != -1) {
			//	The following line only gets executed once per millennium.
			play_sound(IDW_MILDRUM);
		}
		changeMode = 1;
	} else {
		if (ttg >= 1) {
			howmany = (int) ttg;
			which = IDS_DAYS; 
			changeMode = 2;
		} else {
			ttg *= 24;
			if (ttg >= 1) {
				howmany = (int) ttg;
				which = IDS_HOURS;
				changeMode = 3;
			} else {
				ttg *= 60;
				if (ttg >= 1) {
					howmany = (int) ttg;
					which = IDS_MINUTES;
					changeMode = 4;
				} else {
					ttg *= 60;
					howmany = (int) (ttg);
					which = IDS_SECONDS;
					changeMode = 5;
					if (howmany == 0) {
						lstrcpy(s, Format(7));
						goto bailsec;
					}
				}
			}
		}
		if (countToday) {
			howmany++;
		}
		which -= (howmany == 1) ? 1 : 0;
		verb = rstring(howmany == 1 ? IDS_IS : IDS_ARE);
		wsprintf(s, Format(6), (LPSTR) verb, howmany, (LPSTR) rstring(which));
	}
bailsec:;
//sprintf(s, "%.6f ttg=%.6f Rem = %d %s", faketime, ttg, howmany, rstring(which));
	GetTextExtentPoint32(hDC, s, lstrlen(s), &sz);
	fontSize = (40 * (isize * 2)) / (sz.cx * 3);
	SelectObject(hDC, ofont);
	DeleteObject(rfont);
	rfont = CreateFont(fontSize, 0, 0, 0, FW_BOLD, 0, 0, 0, ANSI_CHARSET,
					   OUT_TT_PRECIS, CLIP_TT_ALWAYS, PROOF_QUALITY,
					   0x04 | FF_ROMAN, NULL);
	ofont = SelectObject(hDC, rfont);
	GetTextExtentPoint32(hDC, s, lstrlen(s), &sz);
	SetBkMode(hDC, OPAQUE);
	SetBkColor(hDC, RGB(0, 0, 0));
	SetTextColor(hDC, RGB(255, 255, 255));
	SetTextAlign(hDC, TA_NOUPDATECP);
	TextOut(hDC, ((offX + isize / 2) - (isize / 40)) - (sz.cx / 2),
				 (offY + isize / 2) - (sz.cy / 2), s, lstrlen(s));
	SelectObject(hDC, ofont);
	DeleteObject(rfont);

	ReleaseDC(hWnd, hDC);
}

//	RELEASE_IMAGE  --  Release storage associated with the image

void release_image(void)
{
	if (earthBitmap != NULL) {
		GlobalFree(earthBitmap);
		earthBitmap = NULL;
	}
	if (imagePal != NULL) {
		DeleteObject(imagePal);
		imagePal = NULL;
	}
}
