/*

		Millennium Saver Window procedure
                             
*/

#include "milscrsv.h"

int imageSize = 0;						// Requested image size
int hourChime = TRUE;					// Chime on hour
int halfHourChime = TRUE;				// Chime on half-hour
int playTheme = TRUE;					// Play theme at start ?
int startChime = TRUE;					// Chime at start (if not playing theme) ?
int exitChime = TRUE;					// Chime at exit ?
int showSnake = TRUE;					// Show the snake image ?
int showDateTime = TRUE;				// Show date and time ?
int millenniumStarts = 0;				// Date - 2000 when millennium begins (for compatibility with 1.0)
int targetDate = (int) C2000;			// Date we're counting down to
int countToday = FALSE;					// Count current interval in those remaining
int testMode = FALSE;					// Test mode ?
int themeFileIndex = 0;					// Index of theme sound file in string table

static COLORREF skycolour = RGB(0, 0, 255); // Colour for legends

#ifdef TRACY
void Blooie(char *s)
{
	RECT rc;
	HDC hDC;

	hDC = GetDC(hWndMain);
	GetClientRect(hWndMain, &rc);
	SetBkMode(hDC, OPAQUE);
	SetTextColor(hDC, RGB(255, 255, 255));
	SetBkColor(hDC, RGB(255, 0, 0));
	SetTextAlign(hDC, TA_NOUPDATECP);
	TextOut(hDC, 10, 10, s, strlen(s));
	ReleaseDC(hWndMain, hDC);
}
#endif

/*	PLAY_SOUND  --  Play a designated sound.  Any in-progress sound
					is terminated.  */

void play_sound(int which)
{
	/*	SND_NOSTOP keeps chimes from stepping on the theme if
		we're launched around the time for a chime.  */
	PlaySound(MAKEINTRESOURCE(which), hMainInstance,
		SND_ASYNC | SND_RESOURCE | SND_NOSTOP);
}

//  SET_TM_TIME  --  Set time from Windows system or local time

void set_tm_time(struct tm *t, BOOL islocal)
{
#define CtF(tf, sf) t->tf = s.sf

	if (testMode && faketime > 0) {
		jyear(faketime, &t->tm_year, &t->tm_mon, &t->tm_mday);
		t->tm_year -= 1900;
		t->tm_mon--;
		jhms(faketime, &t->tm_hour, &t->tm_min, &t->tm_sec);
		t->tm_wday = ((long) (faketime + 1.5)) % 7;
	} else {
		SYSTEMTIME s;

		if (islocal) {
			GetLocalTime(&s);
		} else {
			GetSystemTime(&s);
		}
		CtF(tm_sec, wSecond);
		CtF(tm_min, wMinute);
		CtF(tm_hour, wHour);
		CtF(tm_mday, wDay);
		CtF(tm_mon, wMonth - 1);
		CtF(tm_year, wYear - 1900);
		CtF(tm_wday, wDayOfWeek);
		//  tm_yday  never used
		//	tm_isdst never used
//#define BILGEPUMP
#ifdef BILGEPUMP
/*  Back in the SSFRs--"Date did not change at midnight" test case  */
{
	static int bilgeh = 23, bilgem = 59, bilges = 20, bilged = -1;

//t->tm_mday=31; t->tm_mon=11; t->tm_year=99; 
	if (bilged < 0) {
		bilged = t->tm_mday;
	}
	t->tm_mday = bilged;
	t->tm_hour = bilgeh;
	t->tm_min = bilgem;
	t->tm_sec = bilges;
	bilges++;
	if (bilges >= 60) {
		bilges = 0;
		bilgem++;
		if (bilgem >= 60) {
			bilgem = 0;
			bilgeh++;
			if (bilgeh >= 24) {
				bilgeh = 0;
				bilged++;
				//	Just for 00:00 testing, so ignore further propagation
			}
		}
	}
}
#endif
	}
}

/*	FINDTHEMEFILE  --  Find the theme file, searching from the best
					   to the worst fidelity available.  */

void findThemeFile(void)
{
	int f;

	for (f = IDS_WAVEFILE; TRUE; f++) {
		char *fn = rstring(f);
		HFILE hf;
		OFSTRUCT ofs;

		if (fn == NULL || fn[0] == '.') {
			break;
		}
		hf = OpenFile(fn, &ofs, OF_EXIST);
		if (hf != HFILE_ERROR) {
			themeFileIndex = f;
			break;
		}
	}
}

/*  SCREENSAVERPROC  --  Screen saver window procedure.  */

WindowProc(ScreenSaverProc)
{
    HDC hDC;
    PAINTSTRUCT ps;         			// holds PAINT information
    RECT cr;							// Client rectangle
    int minsize;						// Short direction in window
    static int offX, offY;				// Offset of image within window
    static long cctime;					// Time to display (real or simulated) */
    static long rptime = 0;				// Last time image shifted
    struct tm lt;
    static BOOL repositioned = FALSE;
	static char stdname[128], dstname[128];	// Standard and Daylight time zone names
	static int isize = -1;					// Image size
	static TEXTMETRIC tmx;
	static int lastmin;						// Last minutes reading

    switch (Message) {

        case WM_CREATE:
            hDC = GetDC(hWnd);
			SetCursor(NULL);			// Zap cursor before first draw
            GetIniEntries();
            GetIniSettings();			// Get setup defaults
            localtimeformat();
			GetTextMetrics(hDC, &tmx);
            srand((unsigned) (GetTickCount() & 0xFFFF));
            ReleaseDC(hWnd, hDC);
			needImageBitmap();
            SetTimer(hWnd, 1, 1000u, NULL);
			findThemeFile();			// Find theme audio file, if any
			if (!fChildPreview) {
				if (playTheme && themeFileIndex > 0) {
					PlaySound(rstring(themeFileIndex), NULL,
						SND_ASYNC | SND_FILENAME | SND_NOSTOP);
				} else if (startChime) {
					play_sound(IDW_MILDRUM);
				}
			}
			set_tm_time(&lt, TRUE);
			lastmin = lt.tm_min;
            break;

        case WM_DESTROY:
            KillTimer(hWnd, 1);
			release_image();
			if (!fChildPreview && exitChime) {
				PlaySound(MAKEINTRESOURCE(IDW_MILDRUM), hMainInstance,
					SND_RESOURCE);
			}
        	break;

#ifdef CLIPPO

		/*	If CLIPPO is defined, right mouse button copies screen
			to the clipboard.  Why?  So we can make screen shots to
			illustrate the program.  */

		case WM_RBUTTONDOWN:
            {
                HDC hdcMem;
                HBITMAP hBitmap;

                hDC = GetDC(hWnd);
                hdcMem = CreateCompatibleDC(hDC);
				GetClientRect(hWnd, &cr);
                hBitmap = CreateCompatibleBitmap(hDC,
                            cr.right, cr.bottom);
                if (hBitmap) {
                    SelectObject(hdcMem, hBitmap);
                    BitBlt(hdcMem, 0, 0, cr.right, cr.bottom,
                            hDC, 0, 0, SRCCOPY);
                    OpenClipboard(hWnd);
                    EmptyClipboard();
                    SetClipboardData(CF_BITMAP, hBitmap);
                    CloseClipboard();
                    DeleteDC(hdcMem);
                    ReleaseDC(hWnd, hDC);
                }
            }
			return 0;
#endif

		//	Play theme when carriage return is pressed

		case WM_KEYDOWN:
			if (wParam == VK_RETURN && themeFileIndex > 0) {
				PlaySound(rstring(themeFileIndex), NULL,
					SND_ASYNC | SND_FILENAME | SND_NOSTOP);
				return 0;
			}
			break;
        
        case WM_TIMER:
			InvalidateRect(hWnd, NULL, FALSE);
        	break;

		case WM_ERASEBKGND:
			hDC = (HDC) wParam;
		    GetClientRect(hWnd, &cr);
		    FillRect(hDC, &cr, (HBRUSH) GetStockObject(BLACK_BRUSH));
			repositioned = TRUE;
			return 0;

        case WM_PAINT:
			time(&cctime);        
            memset(&ps, 0, sizeof(PAINTSTRUCT));
            hDC = BeginPaint(hWnd, &ps);
			set_tm_time(&lt, TRUE);
			if (testMode && (faketime > 0)) {
				faketime += max(1.0 / (24.0 * 60 * 60), ((targetDate + 0.5) - faketime) / 10);
			} else {
				faketime = jtime(&lt);
			}
            GetClientRect(hWnd, &cr);
            minsize = min(cr.right, cr.bottom);
			isize = (imageSize < 64) ? ((2 * minsize) / 3) : imageSize;
            minsize = min(cr.right, (cr.bottom - (2 * (tmx.tmHeight + 1))));
			if (isize > minsize) {
				isize = minsize;
			}
            
            /* If it's time, randomly reposition the image within the
               rectangular window.  This prevents burning in the phosphor by
               the constant parts of the map. */ 
            
            if ((cctime - rptime) > (RepositionTime * 60L)) {
            	offX = (int) (((cr.right - isize) * ((long) rand())) / RAND_MAX);
            	offY = (int) ((((cr.bottom - (tmx.tmHeight + 1)) - isize) * ((long) rand())) / RAND_MAX);
				offY = max(offY, tmx.tmHeight + 1);            	   
            	rptime = cctime;
            	repositioned = TRUE;
            }

			updviewfrom(hWnd, isize, offX, offY, repositioned);
			repositioned = FALSE;

			//	If requested, chime at appropriate intervals 

			if (!fChildPreview) {
		        cr.left = offX;
		        cr.top = offY - (tmx.tmHeight + 1);
			    cr.right = cr.left + isize;
			    cr.bottom = cr.top + isize + (tmx.tmHeight + 1) * 2;
		       	ssavetime(hDC, lt, cr, skycolour);	// Update local and universal time
				if (!testMode) {		//	Chimes are irritating in test mode
					if (hourChime && (lastmin > lt.tm_min)) {
						play_sound(IDW_MILDRUM);
					}
					if (halfHourChime && (lastmin < 30 && lt.tm_min >= 30)) {
						play_sound(IDW_MILDRUM1);
					}
				}
				lastmin = lt.tm_min;
			}

            ValidateRgn(hWnd, NULL);

            /* Inform Windows painting is complete */
            EndPaint(hWnd, &ps);
            return 0;
    }
	return DefScreenSaverProc(hWnd, Message, wParam, lParam);
}

